import math
from cv2.typing import MatLike
from typing import Tuple
import cv2
import numpy as np

from cardscan.image_processing import copy_bgr


def grid_layout_tile_size(tile_ratio, tile_count, window_size):
    window_w, window_h = window_size
    new_tile_width = new_tile_height = rows = cols = -1
    for cols in range(1, 30):
        new_tile_width = window_w // cols
        new_tile_height = int(new_tile_width / tile_ratio)
        rows = math.ceil(tile_count / cols)
        if rows * new_tile_height <= window_h:
            break
    return (new_tile_width, new_tile_height), (rows, cols)


def scaleWidthKeepRatio(frame: MatLike, bounding_size: Tuple[int, int]) -> MatLike:
    bound_w, bound_h = bounding_size
    frame_h, frame_w = frame.shape[:2]
    scale_factor_fit_w = bound_w / frame_w
    scale_factor_fit_h = bound_h / frame_h
    scale_factor = min(scale_factor_fit_w, scale_factor_fit_h)
    resized_frame = cv2.resize(
        frame,
        (int(frame_w * scale_factor), int(frame_h * scale_factor)),
        interpolation=cv2.INTER_CUBIC,
    )
    return resized_frame


def get_top_left(i: int, cols: int, size: Tuple[int, int]):
    col_i = i % cols
    row_i = i // cols
    return col_i * size[0], row_i * size[1]


def create_tile_gallery(
    tiles, img_ratio: float, window_size: Tuple[int, int]
) -> MatLike:
    tile_bounding_size, (tot_rows, tot_cols) = grid_layout_tile_size(
        img_ratio, len(tiles), window_size
    )

    tiles = [scaleWidthKeepRatio(tile, tile_bounding_size) for tile in tiles]
    tiles_top_left = [
        get_top_left(i, tot_cols, tile_bounding_size) for i, tile in enumerate(tiles)
    ]

    window_w, window_h = window_size
    img_shape = (window_h, window_w, 3)
    img = np.zeros(img_shape, dtype=np.uint8)

    for tile, (x, y) in zip(tiles, tiles_top_left):
        h, w = tile.shape[:2]
        img[y : y + h, x : x + w] = copy_bgr(tile)
    return img



from cv2 import transform
from cardscan.pipeline_instances.rect_perspective_pipeline import *


def draw_contours(contours):
    img = card_contours_transform.initial_input
    if contours is None:
        return img.copy()
    img = copy_bgr(img)
    return [cv2.drawContours(img, contours, -1, (0, 255, 255), 3)]


def draw_contours_def(contours_def):
    c, h = contours_def
    return draw_contours(c)


def debug_image(img, *_):
    return [img]


def debug_images(imgs, *_):
    return imgs


transform_to_img = {
    inner_unique_contours_transform: draw_contours,
    find_contours_def: draw_contours_def,
    approx_filter_quad: draw_contours,
    filter_contours_by_size_solidity: draw_contours,
    card_contours_transform: draw_contours,
    copy_gray: debug_image,
    canny_transform: debug_image,
    rotate_top_left_corner_low_density_transform: debug_images,
    selective_adaptative_threshold: debug_images,
    card_images: debug_images,
    perspective_crop_contours: debug_images,
    close_range_card_contours_detection: draw_contours,
    long_range_card_contours_detection: draw_contours,
    card_contours_detection: draw_contours,
}


def get_debug_imgs(results: List, keep_results: List):
    tiles = []
    for output, transform in zip(results, keep_results):
        imgs = transform_to_img[transform](output)
        # TODO clear err message if transform not found
        for img in imgs:
            tiles.append(img)
    return tiles



import cv2
from cv2.typing import MatLike


def contour_to_poly(contour):
    """Approximate the contour to a polygon."""
    epsilon = 0.04 * cv2.arcLength(contour, True)  # Adjust epsilon as needed
    approx_polygon = cv2.approxPolyDP(contour, epsilon, True)
    return approx_polygon


def contours_to_poly(contours):
    """Approximate the contours to polygones"""
    return [contour_to_poly(contour) for contour in contours]


def contour_approx_poly_filter_quadrilater(contours):
    return filter_4_edges_contours(contours_to_poly(contours))
    # convex_contours = [cv2.convexHull(contour) for contour in contours]


# See https://docs.opencv.org/4.8.0/d1/d32/tutorial_py_contour_properties.html
def calc_solidity(contour):
    area = cv2.contourArea(contour)
    hull = cv2.convexHull(contour)
    hull_area = cv2.contourArea(hull)
    solidity = float(area) / hull_area
    return solidity


def filter_4_edges_contours(contours):
    return [c for c in contours if len(c) is 4]


def filter_contours_by_size_solidity(contours, min_solidity: float = 0.9):
    return [
        contour
        for contour in contours
        if cv2.contourArea(contour) > 100 and calc_solidity(contour) > min_solidity
    ]



from cardscan.pipeline import Pipeline, PipelineChoice
import cv2

from cardscan.image_processing import *
from cardscan.image_processing.contour_hierarchy import inner_unique_contours

inner_unique_contours_transform = lambda contours_def: inner_unique_contours(
    contours_def[1], contours_def[0]
)

find_contours_def = lambda img: cv2.findContours(
    img, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE
)


def approx_filter_quad(contours):
    return filter_4_edges_contours(contours_to_poly(contours))


card_contours_transform = Pipeline(
    (
        find_contours_def,
        inner_unique_contours_transform,
        filter_contours_by_size_solidity,
        approx_filter_quad,
    )
)


def selective_adaptative_threshold(img, *_):
    """Adaptative threshold applied on the non-dark area - black otherwise - to avoid holes in the borders if they are too thick."""

    # Define the size of the local neighborhood for calculating the mean
    block_size = 5

    # Calculate the mean of the local neighborhood for each pixel
    mean_values = cv2.boxFilter(img, ddepth=-1, ksize=(block_size, block_size))

    # Define a custom threshold value below which pixels become black
    custom_threshold = 85

    # Create a mask to identify pixels where the mean value is below the custom threshold
    custom_mask = mean_values < custom_threshold

    # Apply adaptive thresholding only to pixels where the custom mask is False
    adaptive_thresholded = cv2.adaptiveThreshold(
        img, 255, cv2.ADAPTIVE_THRESH_MEAN_C, cv2.THRESH_BINARY, 9, 20
    )

    # TODO compare and benchmark this other way to do it
    # adaptive_thresholded = cv2.adaptiveThreshold(
    #    img, 255, cv2.ADAPTIVE_THRESH_GAUSSIAN_C, cv2.THRESH_BINARY, 31, 30
    # )

    # Combine the adaptive thresholding result with the custom thresholding result
    result = np.where(custom_mask, 0, adaptive_thresholded)
    return result


def perspective_crop_transform(contours, img):
    return [perspective_crop(contour, img) for contour in contours]


rotate_top_left_corner_low_density_transform = lambda imgs: [
    rotate_top_left_corner_low_density(img, corner_width_ratio=0.1) for img in imgs
]


def canny_transform(img):
    return cv2.Canny(img, threshold1=150, threshold2=255, apertureSize=3)


card_images = Pipeline()


def gaussian_transform(img):
    return cv2.GaussianBlur(img, (5, 5), 1)


close_range_card_contours_detection = (canny_transform, card_contours_transform)
long_range_card_contours_detection = (
    selective_adaptative_threshold,
    card_contours_transform,
)

card_contours_detection = card_images.choice(
    close_range_card_contours_detection,
    long_range_card_contours_detection,
    first_where=len,
)


perspective_crop_contours = card_images.bind_input(perspective_crop_transform)
card_images.setTransforms(
    (
        copy_gray,
        card_contours_detection,
        perspective_crop_contours,
        rotate_top_left_corner_low_density_transform,
    )
)


# The blur option help smoth artefact when detecting from a compressed image
def scan(
    *args,
    blur: bool = False,
    **kwargs,
):
    if blur:
        card_images.transforms = (gaussian_transform,) + card_images.transforms
    return card_images.run(*args, **kwargs)



import numpy as np
from cv2.typing import MatLike
import cv2
from typing import List, Tuple

Contour = List[Tuple[int, int]]


def perspective_crops(contours: List[Contour], img: MatLike):
    return [perspective_crop(contour, img) for contour in contours]


def perspective_crop(contour: Contour, img: MatLike, shape=(200, 200)):
    src_points = contour

    h, w = shape[:2]
    top_left_coord = [0, 0]
    top_right_coord = [w, 0]
    bottom_left_coord = [0, h]
    bottom_right_coord = [w, h]

    src_points = np.array(src_points, dtype=np.float32)
    dst_points = np.array(
        [
            top_right_coord,
            top_left_coord,
            bottom_left_coord,
            bottom_right_coord,
        ],
        dtype=np.float32,
    )
    transformation_matrix = cv2.getPerspectiveTransform(src_points, dst_points)
    transformed_image = cv2.warpPerspective(img, transformation_matrix, (w, h))
    cropped_img = transformed_image[0:h, 0:w]
    return cropped_img


def rotate_top_left_corner_low_density(img: MatLike, corner_width_ratio: float = 0.1):
    """Compare the mean intensity of the 4 corners of `img` where each corner is a square with size corner_width_ratio*img_width. Rotate the image so the corner with lowest intensity (more black) is located at the top left."""

    h, w = img.shape[:2]
    corner_size = int(corner_width_ratio * w)
    corners_topleft = [
        (0, 0),  # Top-left corner
        (img.shape[1] - corner_size, 0),  # Top-right corner
        (0, img.shape[0] - corner_size),  # Bottom-left corner
        (img.shape[0] - corner_size, img.shape[1] - corner_size),  # Bottom-right corner
    ]
    """The top-left position of corners in order TL-TR-BR-BL"""

    squares = []
    """Corners subimage of the image in respectve order to `corners_topleft`."""

    mean_intensities = []
    """Mean intensity of each corner in respectve order to `corners_topleft`"""

    for corner_x, corner_y in corners_topleft:
        square = img[
            corner_x : corner_x + corner_size, corner_y : corner_y + corner_size
        ]
        mean_intensity = np.mean(square)
        squares.append(square)
        mean_intensities.append(mean_intensity)

    min_intensity_index = np.argmin(mean_intensities)
    """The index of the corner with lowest mean intensity"""

    angles = [0, -90, 90, 180]
    """The rotation angle needed to rotate the image so the index associated corner is located at the top-left of the image. In respectve order to `corners_topleft`."""

    angle = angles[min_intensity_index]
    """ The rotation angle to align the corner with the lowest mean intensity to the top-left"""

    if angle == 0:
        rotated_img = img
    else:
        rows, cols, *_ = img.shape
        M = cv2.getRotationMatrix2D((cols / 2, rows / 2), angle, 1)
        rotated_img = cv2.warpAffine(img, M, (cols, rows))

    return rotated_img



import cv2
from cv2.typing import MatLike


def has_3_channels(img: MatLike) -> bool:
    return img.shape[-1] == 3


def copy_rgb(img: MatLike) -> MatLike:
    """Convert from BGR or GRAY ro RGB"""
    if has_3_channels(img):
        return cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
    else:
        return cv2.cvtColor(img, cv2.COLOR_GRAY2RGB)


def copy_bgr(img: MatLike) -> MatLike:
    """If image is 3 channel, just copy. Else, convert from GRAY to BGR"""
    if has_3_channels(img):
        return img.copy()
    return cv2.cvtColor(img, cv2.COLOR_GRAY2BGR)


def copy_gray(img: MatLike) -> MatLike:
    """If image is 3 channel, convert from GRAU to BGR. Else, just copy."""
    if has_3_channels(img):
        return cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    return img.copy()



from typing import Callable, List, Optional, Any, Type

from cv2.typing import MatLike

import copy


class PipelineChoice:
    def __init__(self, *choices, where):
        self.choices = choices
        self.where = where

    def run(self, input, keep_results, aggregated_results):
        output = None
        for choice in self.choices:
            input_copy = copy.deepcopy(input)
            output = Pipeline._run(input_copy, choice, keep_results, aggregated_results)
            if self.where(output):
                break
        return output


# class BindInput:


Transform = Callable | List | PipelineChoice


class Pipeline:
    initial_input_getter = lambda: None

    def __init__(
        self,
        transforms: List[Transform] | None = None,
    ):
        self.transforms = transforms
        self.initial_input = None
        self.aggregated_results = {}
        self.keep_results = []

    def choice(self, *choices, first_where):
        """Choose the first pipeline which output passes the `first_where` test function."""

        def conditionalPipeline(input):
            output = None
            for choice in choices:
                input_copy = copy.deepcopy(input)
                output = self._run(input_copy, choice)
                if first_where(output):
                    break
            return output

        return conditionalPipeline

    def bind_input(self, fn: Callable) -> Callable:
        """Bind the pipeline input as last parameter of `fn`."""
        return lambda *args, **kwargs: fn(*args, self.get_input(), **kwargs)

    def setTransforms(
        self,
        transforms: List[Transform],
    ):
        self.transforms = transforms

    def get_input(self):
        return self.initial_input

    def _run(self, input, transform):
        if isinstance(transform, Callable):
            input = transform(input)
        elif isinstance(transform, Pipeline):
            sub_pipeline = transform
            sub_pipeline.aggregated_results = self.aggregated_results
            sub_pipeline.keep_results = self.keep_results
            sub_pipeline.initial_input = input
            input = sub_pipeline._run(input, sub_pipeline.transforms)

        elif isinstance(transform, tuple):
            for subtransform in transform:
                input = self._run(
                    input,
                    subtransform,
                )
        else:
            # TODO
            print("ERROR unknow type")

        if transform in self.keep_results:
            self.aggregated_results[transform] = copy.deepcopy(input)
        return input

    def run(
        self,
        input: Any,
        keep_results: List[Transform] = [],
    ):
        self.aggregated_results = {}
        self.keep_results = keep_results
        self.initial_input = input

        if len(keep_results) is 0:
            return self._run(input, self.transforms)

        output = []
        for item in keep_results:
            self.aggregated_results[item] = []
        self._run(input, self.transforms)
        if self in keep_results:
            self.aggregated_results[self] = copy.deepcopy(input)
        for item in keep_results:
            output.append(self.aggregated_results[item])
        return output



import sys
import cv2
from cardscan.pipeline import Pipeline
from cardscan.image_processing import *
from typing import Callable

from cardscan import (
    scan,
    card_contours_transform,
    rotate_top_left_corner_low_density_transform,
)
from cardscan.pipeline_instances import transform_to_img


def run_cam_app(fn: Callable):
    cam = cv2.VideoCapture(0)
    window_name = "Track users"
    cv2.namedWindow(window_name, cv2.WND_PROP_FULLSCREEN)
    cv2.setWindowProperty(window_name, cv2.WND_PROP_FULLSCREEN, cv2.WINDOW_FULLSCREEN)

    while True:
        window_rect = cv2.getWindowImageRect(window_name)
        window_width = window_rect[2]
        window_height = window_rect[3]
        window_size = (window_width, window_height)

        ret, frame = cam.read()
        # frame = cv2.flip(frame, 1)
        if not ret:
            break

        img = fn(frame, window_size)
        cv2.imshow(window_name, img)

        key = cv2.waitKey(1)
        ESC_KEY = 27
        if key == ESC_KEY or key == ord("q"):
            break

    cam.release()
    cv2.destroyAllWindows()
    sys.exit()



from cv2.typing import MatLike
from typing import Tuple, List

from cardscan.pipeline import Pipeline
from cardscan.image_processing import *

from cardscan.pipeline_instances.display_transforms import get_debug_imgs

from .tile_gallery import create_tile_gallery
from .cam_app import run_cam_app


def run_pip_app(pipeline: Pipeline, keep_results: List):
    def get_screen_img(frame: MatLike, window_size: Tuple[int, int]):
        results = pipeline.run(frame, keep_results=keep_results)
        tiles = get_debug_imgs(results, keep_results)

        (h, w) = frame.shape[:2]
        tile_ratio = w / h
        img = create_tile_gallery(tiles, tile_ratio, window_size)
        return img

    run_cam_app(get_screen_img)



from typing import List


class Hierarchy:
    NEXT = 0
    PREVIOUS = 1
    FIRST_CHILD = 2
    PARENT = 3
    OUTER = 0
    INNER = 1
    BLACK_OUTER = WHITE_INNER = 1
    WHITE_OUTER = BLACK_INNER = 0

    def __init__(self, h):
        self.h = h[0]

    def black_outer_entries(self):
        entries = []
        for child in self.siblings_from_first(0):
            for grandchild in self.children(child):
                entries.append(grandchild)
        return entries

    def next(self, i):
        return self.h[i][self.NEXT]

    def parent(self, i):
        return self.h[i][self.PARENT]

    def first_child(self, i: int):
        return self.h[i][self.FIRST_CHILD]

    def siblings_from_first(self, i: int) -> List[int]:
        siblings = []
        while i != -1:
            siblings.append(i)
            i = self.next(i)
        return siblings

    def children(self, i: int) -> List[int]:
        first_child = self.first_child(i)
        return self.siblings_from_first(first_child)

    def grandchildren(self, i: int) -> List[int]:
        grandchildren = []
        for child in self.children(i):
            for grandchild in self.children(child):
                grandchildren.append(grandchild)
        return grandchildren


def _aggreate_inner_unique_contours(
    hier: Hierarchy, outer_contours: List[int], aggregated_results=[]
):
    for outer_contour in outer_contours:
        inner_contours = hier.children(outer_contour)
        child_count = len(inner_contours)
        if child_count == 1:
            aggregated_results.append(inner_contours[0])
        elif child_count:
            nested_outer_contours = []
            for inner_contour in inner_contours:
                nested_outer_contours += hier.children(inner_contour)
            _aggreate_inner_unique_contours(
                hier, nested_outer_contours, aggregated_results
            )


def inner_unique_contours(h, c):
    if h is None or len(c) == 0:
        return []
    hier = Hierarchy(h)
    i_results = []
    _aggreate_inner_unique_contours(hier, hier.black_outer_entries(), i_results)
    contour_results = [c[i] for i in i_results]
    return contour_results



import cv2


def getCv2WindowSize(window_name: str):
    window_rect = cv2.getWindowImageRect(window_name)
    window_width = window_rect[2]
    window_height = window_rect[3]
    window_size = (window_width, window_height)
